"""Service to use email verification client with database."""
import logging
from typing import Dict, List, Tuple

import requests

from email_verification.client import EmailVerificationClient
from email_verification.database import ResultDatabase
from email_verification.utils.response_normilizer import response_normalizer


class EmailVerificationService(object):
    """Email verification service."""

    def __init__(self, api_key: str, db: ResultDatabase, timeout: int = 2) -> None:
        """Initialize the service."""
        self.client = EmailVerificationClient(api_key, timeout)
        self.db = db
        self.logger = logging.getLogger(__name__)

    def count_emails(self, domain: str, **kwargs) -> Dict[str, str]:
        """Count the number of emails in the given domain and handle errors."""
        try:
            return self.client.count_emails(domain, **kwargs)
        except requests.exceptions.RequestException as request_exception:
            self.logger.error('Request error for domain {0}: {1}'.format(domain, request_exception))
            return {'Error': 'Request error: {request_exception}'.format(request_exception=str(request_exception))}
        except Exception as exception:
            self.logger.error('Unexpected error for domain {0}: {1}'.format(domain, exception))
            return {'Error': 'Unexpected error: {exception}'.format(exception=str(exception))}

    def perform_verification(self, email: str) -> Dict[str: str]:
        """Email verification service using EmailVerificationClient and database."""
        try:
            verification_result = self.client.verify_email(email)
        except requests.exceptions.RequestException as request_exception:
            self.logger.error('Request error for email {0}: {1}'.format(email, request_exception))
            return email, {'Request error': '{request_exception}'.format(request_exception=str(request_exception))}
        except Exception as exception:
            self.logger.error('Unexpected error for email {0}: {1}'.format(email, exception))
            return email, {'Unexpected error': '{exception}'.format(exception=str(exception))}
        verification_result = response_normalizer(verification_result)
        email_to_save = verification_result.pop('meta_params_email')
        self.db.create_result(email_to_save, verification_result)
        return verification_result

    def get_all_results(self) -> List[Dict[str, str]]:
        """Show all verification results."""
        return self.db.get_results()

    def delete_result(self, email: str) -> Tuple[str, Dict]:
        """Delete result and handle errors."""
        try:
            self.db.delete_result(email)
        except Exception as exception:
            return email, {'error': 'Unexpected error: {exception}'.format(exception=str(exception))}
        return email, {'status': 'deleted'}

    def get_result(self, email: str) -> Tuple[str, Dict]:
        """Get verification email if exists."""
        try:
            get_result = self.db.get_result(email)
        except Exception as exception:
            return email, {'error': 'Unexpected error: {exception}'.format(exception=str(exception))}
        return email, get_result



"""Verify email using hunter."""
from typing import Dict, Optional
from urllib.parse import urlencode, urljoin

import requests


class EmailVerificationClient(object):
    """Verify email using hunter."""

    http_ok = 200

    def __init__(self, api_key: str, timeout: int = 2) -> None:
        """Initialize the client."""
        self.api_key = api_key
        self.timeout = timeout
        self.base_url = 'https://api.hunter.io/v2/'

    def verify_email(self, email: str) -> Dict[str, str]:
        """Verify email using hunter."""
        request_params = {'email': email}
        url = self._url_builder('email-verifier', request_params)
        response_data = self._make_request(url)
        return response_data.json()

    def count_emails(self, domain: str, company: str = None, email_type: str = None):
        """Count all emails using hunter."""
        request_params = {'domain': domain, 'company': company, 'email_type': email_type}
        url = self._url_builder('email-count', request_params)
        response_data = self._make_request(url)
        return response_data.json()

    def _url_builder(self, url: str, request_params: Dict[str, Optional[str]]) -> str:
        """Build url using urllib and format."""
        request_params['api_key'] = self.api_key
        request_params = {key: entered_param for key, entered_param in request_params.items() if entered_param}
        request_params = urlencode(request_params)
        endpoint = '{request_url}?{request_params}'.format(request_url=url, request_params=request_params)
        return urljoin(self.base_url, endpoint)

    def _make_request(self, request_url: str) -> requests.Response:
        """Make a request to hunters."""
        response = requests.get(request_url, timeout=self.timeout)
        if response.status_code != self.http_ok:
            response.raise_for_status()
        return response



# setup.py
from setuptools import setup, find_packages

setup(
    name='email_verification',
    version='0.4.0',
    packages=find_packages(),
    install_requires=[
        'requests>=2',
        'urllib3==2.1.0',

    ],
)



"""Database."""
from typing import Dict, List


class ResultDatabase(object):
    """Base class with CRUD operations."""

    def __init__(self) -> None:
        """Initialize the database."""
        self.checked_emails: List[Dict[str, str]] = []

    def create_result(self, email: str, verification_result: Dict[str, str]) -> bool:
        """Create if not exist."""
        if self.exists(email):
            self.update_result(email, verification_result)
            return False
        verification_result['email'] = email.upper()
        self.checked_emails.append(verification_result)
        return True

    def get_results(self) -> List[Dict[str, str]]:
        """Get all results."""
        return self.checked_emails

    def get_result(self, email: str) -> Dict[str, str]:
        """Get result for a specific email."""
        for inner_result in self.checked_emails:
            if inner_result['email'].upper() == email.upper():
                return inner_result
        return {'error': 'Result not found for email: {email}'.format(email=email)}

    def delete_result(self, email: str) -> None:
        """Delete if exists."""
        new_checked_emails = []
        for inner_result in self.checked_emails:
            if inner_result['email'].upper() != email.upper():
                new_checked_emails.append(inner_result)
        self.checked_emails = new_checked_emails

    def update_result(self, email: str, verification_result: Dict[str, str]) -> None:
        """Update if exists."""
        for inner_email in self.checked_emails:
            if inner_email['email'] == email.upper():
                inner_email.update(verification_result)

    def exists(self, email: str) -> bool:
        """Check if email exists."""
        return any(inner_email['email'] == email.upper() for inner_email in self.checked_emails)



"""Response Normalizer."""
from typing import Dict


def response_normalizer(nested_dict: Dict, parent_key: str = '', sep: str = '_') -> Dict[str, str]:
    """Response Normalizer."""
    flattened_dict = {}
    for key, inner_value in nested_dict.items():
        new_key = '{parent_key}{sep}{key}'.format(parent_key=parent_key, sep=sep, key=key) if parent_key else key
        if isinstance(inner_value, dict):
            flattened_dict.update(response_normalizer(inner_value, new_key, sep=sep))
        else:
            flattened_dict[new_key] = inner_value
    return flattened_dict
